try:
    import unittest2
except ImportError:
    import unittest as unittest2
import sys
import os
from os import path
from argparse import Namespace
import subprocess
import Bio.SearchIO
from minimock import mock, restore, Mock, TraceTracker, assert_same_trace
from antismash import utils
from antismash import config
from Bio.SeqFeature import FeatureLocation


class FakeRecord(object):
    def __init__(self, features=None, seq='FAKESEQ'):
        if features is None:
            features = []
        self.features = features
        self.seq = FakeSeq(seq)

class FakeFeature(object):
    def __init__(self, type_, location=None):
        self.type = type_
        self.qualifiers = {}
        self.location = location

    def extract(self, seq):
        return seq

    def __repr__(self):
        return "FakeFeature(%r, %r, %r)" % (self.location, self.type,
                                            self.qualifiers)

class FakeSeq(object):
    def __init__(self, seq):
        self.seq = seq

    def translate(self, to_stop):
        return self.seq

    def __str__(self):
        return self.seq


class TestUtils(unittest2.TestCase):

    def setUp(self):
        self.rec = FakeRecord()
        self.types = ["CDS", "gene", "CDS", "CDS_motif", "cluster", "CDS"]
        idx = 1
        for t in self.types:
            f = FakeFeature(t)
            f.qualifiers['locus_tag'] = [ "orf%04d" % idx]
            f.qualifiers['note'] = [ 'PFAM-Id: PF%05d' % idx ]
            f.qualifiers['translation'] = ['FAKESEQ']
            self.rec.features.append(f)
            idx += 1


    def test_get_all_features_of_type(self):
        "Test utils.get_all_features_of_type()"
        for t in set(self.types):
            f = utils.get_all_features_of_type(self.rec, t)
            self.assertEqual(len(f), self.types.count(t))


    def test_get_cds_features(self):
        "Test utils.get_all_cds_features()"
        cds = utils.get_cds_features(self.rec)
        features = utils.get_all_features_of_type(self.rec, "CDS")
        self.assertListEqual(cds, features)


    def test_get_cluster_features(self):
        "Test utils.get_cluster_features()"
        clusters = utils.get_cluster_features(self.rec)
        features = utils.get_all_features_of_type(self.rec, "cluster")
        self.assertListEqual(clusters, features)

    def test_get_pfam_features(self):
        "Test utils.get_pfam_features()"
        motifs = utils.get_pfam_features(self.rec)
        features = utils.get_all_features_of_type(self.rec, "CDS_motif")
        self.assertListEqual(motifs, features)

    def test_get_feature_dict(self):
        "Test utils.get_feature_dict()"
        fd = utils.get_feature_dict(self.rec)
        ids = [f.qualifiers['locus_tag'][0] for f in self.rec.features if f.type == "CDS"]
        keys = fd.keys()
        ids.sort()
        keys.sort()
        self.assertListEqual(ids, keys)

    def test_get_multifasta(self):
        "Test utils.get_multifasta"
        expected = """>orf0001
FAKESEQ
>orf0003
FAKESEQ
>orf0006
FAKESEQ"""

        ret = utils.get_multifasta(self.rec)
        self.assertMultiLineEqual(expected, ret)


class TestContainmentHelpers(unittest2.TestCase):
    def setUp(self):
        self.features = [
            FakeFeature('cluster', FeatureLocation(25, 50)),
            FakeFeature('CDS', FeatureLocation(15, 20)),
            FakeFeature('CDS_motif', FeatureLocation(15, 17)),
            FakeFeature('CDS', FeatureLocation(23, 42)),
            FakeFeature('CDS', FeatureLocation(45, 47)),
            FakeFeature('CDS', FeatureLocation(48, 55)),
            FakeFeature('CDS', FeatureLocation(4700, 4710)),
            FakeFeature('CDS', FeatureLocation(4750, 4760)),
            FakeFeature('CDS', FeatureLocation(4790, 4812)),
            FakeFeature('cluster', FeatureLocation(4711, 4800)),
        ]

        self.record = FakeRecord(self.features)


    def test_get_cluster_cds_features(self):
        "Test utils.get_cluster_cds_features()"
        cluster1, cluster2 = utils.get_cluster_features(self.record)
        self.assertEqual(self.features[0], cluster1)
        self.assertEqual(self.features[-1], cluster2)

        clusterfeatures = utils.get_cluster_cds_features(cluster1, self.record)
        self.assertEqual(self.features[3:6], clusterfeatures)

        clusterfeatures = utils.get_cluster_cds_features(cluster2, self.record)
        self.assertEqual(self.features[-3:-1], clusterfeatures)


    def test_get_withincluster_cds_features(self):
        "Test utils.get_withincluster_cds_features()"
        features = utils.get_withincluster_cds_features(self.record)
        self.assertEqual(self.features[3:6] + self.features[-3:-1], features)


    def test_features_overlap(self):
        "Test utils.features_overlap()"
        self.assertFalse(utils.features_overlap(self.features[0], self.features[1]))
        self.assertFalse(utils.features_overlap(self.features[1], self.features[0]))

        self.assertTrue(utils.features_overlap(self.features[0],  self.features[3]))
        self.assertTrue(utils.features_overlap(self.features[3],  self.features[0]))

        self.assertTrue(utils.features_overlap(self.features[0],  self.features[4]))
        self.assertTrue(utils.features_overlap(self.features[4],  self.features[0]))

        self.assertTrue(utils.features_overlap(self.features[0],  self.features[5]))
        self.assertTrue(utils.features_overlap(self.features[5],  self.features[0]))


    def check_sorting(self):
        for i in range(len(self.features) - 1):
            if self.features[i].location.start > \
               self.features[i+1].location.start:
                return False
            if self.features[i].location.start == \
               self.features[i+1].location.start and \
               self.features[i].location.end > \
               self.features[i+1].location.end:
                return False

        return True

    def test_sort_features(self):
        "Test utils.sort_features()"

        self.assertFalse(self.check_sorting())

        utils.sort_features(self.record)

        self.assertTrue(self.check_sorting())


    def test_cmp_feature_location(self):
        "Test utils.cmp_feature_location()"
        self.assertEqual(0, utils.cmp_feature_location(self.features[0], self.features[0]))
        self.assertEqual(-1, utils.cmp_feature_location(self.features[1], self.features[0]))
        self.assertEqual(1, utils.cmp_feature_location(self.features[0], self.features[1]))


class TestExecutableHelpers(unittest2.TestCase):
    def setUp(self):
        self.config = Namespace()
        self.config.cpus = 2
        config.set_config(self.config)
        self.tt = TraceTracker()
        proc = Mock('proc', tracker=self.tt, returncode=0)
        proc.communicate = Mock('proc.communicate', returns=('output', 'error'), tracker=self.tt)
        mock('subprocess.Popen', tracker=self.tt, returns=proc)

    def tearDown(self):
        restore()

    def test_locate_executable(self):
        "Test utils.locate_executable()"
        test_executable = sys.argv[0]
        if not path.isfile(test_executable):
            self.skip("%r is not a file, skipping test" % test_executable)

        if not os.access(test_executable, os.X_OK):
            self.skip("%r is not executable, skipping test" % test_executable)

        self.assertEqual(utils.locate_executable(test_executable),
                         test_executable)

        discard, short_path = path.split(test_executable)
        self.assertEqual(utils.locate_executable(short_path), test_executable)

        self.assertIsNone(utils.locate_executable("totallymaedup"))


    def test_locate_file(self):
        "Test utils.locate_file()"
        self.assertEqual(__file__, utils.locate_file(__file__))
        self.assertIsNone(utils.locate_file('totallymadeup'))


    def test_excute_no_input(self):
        "Test utils.execute() without stdin input"
        expected = """    Called subprocess.Popen(
        ['fake', '--with', 'parameters'],
        stderr=-1,
        stdin=None,
        stdout=-1)
    Called proc.communicate(input=None)"""
        cmd = ['fake', '--with', 'parameters']

        out, err, retcode = utils.execute(cmd)

        self.assertEqual('output', out)
        self.assertEqual('error', err)
        self.assertEqual(0, retcode)
        assert_same_trace(self.tt, expected)


    def test_execute_with_input(self):
        "Test utils.execute() with stdin input"
        expected = """    Called subprocess.Popen(
        ['fake', '--with', 'parameters'],
        stderr=-1,
        stdin=-1,
        stdout=-1)
    Called proc.communicate(input='fake input')"""
        cmd = ['fake', '--with', 'parameters']

        out, err, retcode = utils.execute(cmd, input='fake input')

        self.assertEqual('output', out)
        self.assertEqual('error', err)
        self.assertEqual(0, retcode)
        assert_same_trace(self.tt, expected)


    def test_run_hmmsearch(self):
        "Test utils.run_hmmsearch()"
        mock('Bio.SearchIO.parse', tracker=self.tt, returns=['mock result'])
        mock('utils.execute', tracker=self.tt, returns=('output', 'error', 0))

        expected = r"""    Called utils.execute(
        ['hmmsearch', '--cpu', '2', 'fake.hmm', '-'],
        input='>testinput\nMADEUP')
    Called Bio.SearchIO.parse(
        <cStringIO.StringI object at ...>,
        'hmmer3-text')"""

        hits = utils.run_hmmsearch('fake.hmm', ">testinput\nMADEUP")
        self.assertEqual(len(hits), 1)

        hit = hits.pop()
        self.assertEqual('mock result', hit)
        assert_same_trace(self.tt, expected)


    def test_run_hmmscan(self):
        "Test utils.run_hmmscan()"
        mock('Bio.SearchIO.parse', tracker=self.tt, returns=['mock result'])
        mock('utils.execute', tracker=self.tt, returns=('output', 'error', 0))

        expected = r"""    Called utils.execute(
        ['hmmscan', '--cpu', '2', 'fake.hmm', '-'],
        input='>testinput\nMADEUP')
    Called Bio.SearchIO.parse(
        <cStringIO.StringI object at ...>,
        'hmmer3-text')"""

        hits = utils.run_hmmscan('fake.hmm', ">testinput\nMADEUP")
        self.assertEqual(len(hits), 1)

        hit = hits.pop()
        self.assertEqual('mock result', hit)
        assert_same_trace(self.tt, expected)


    def test_run_hmmpfam2(self):
        "Test utils.run_hmmpfam2()"
        mock('Bio.SearchIO.parse', tracker=self.tt, returns=['mock result'])
        mock('utils.execute', tracker=self.tt, returns=('output', 'error', 0))

        expected = r"""    Called utils.execute(
        ['hmmpfam2', '--cpu', '2', 'fake.hmm', '-'],
        input='>testinput\nMADEUP')
    Called Bio.SearchIO.parse(
        <cStringIO.StringI object at ...>,
        'hmmer2-text')"""

        hits = utils.run_hmmpfam2('fake.hmm', ">testinput\nMADEUP")
        self.assertEqual(len(hits), 1)

        hit = hits.pop()
        self.assertEqual('mock result', hit)
        assert_same_trace(self.tt, expected)


    def test_get_full_path(self):
        "Test utils.get_full_path()"
        expected = path.join(path.dirname(path.abspath(__file__)), 'fake')
        ret = utils.get_full_path(__file__, 'fake')

        self.assertEqual(ret, expected)


class TestAnalysisHelpers(unittest2.TestCase):
    def test_get_gene_id_locus_tag(self):
        "Test utils.get_gene_id() with locus tag"
        expected = 'test_tag'
        f = FakeFeature("CDS")
        f.qualifiers['locus_tag'] = [expected]

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)


    def test_get_gene_id_gene(self):
        "Test utils.get_gene_id() with gene tag"
        expected = 'test_gene'
        f = FakeFeature("CDS")
        f.qualifiers['gene'] = [expected]

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)


    def test_get_gene_id_protein_id(self):
        "Test utils.get_gene_id() with protein_id tag"
        expected = 'test_id'
        f = FakeFeature("CDS")
        f.qualifiers['protein_id'] = [expected]

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)

    def test_get_gene_id_locus_tag_gene(self):
        "Test utils.get_gene_id() with locus tag and gene tag"
        expected = 'test_tag'
        f = FakeFeature("CDS")
        f.qualifiers['locus_tag'] = [expected]
        f.qualifiers['gene'] = ['test_gene']

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)

    def test_get_gene_id_locus_tag_protein(self):
        "Test utils.get_gene_id() with locus tag and protein_id tag"
        expected = 'test_tag'
        f = FakeFeature("CDS")
        f.qualifiers['locus_tag'] = [expected]
        f.qualifiers['protein_id'] = ['test_id']

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)

    def test_get_gene_id_gene_protein(self):
        "Test utils.get_gene_id() with gene tag and protein_id tag"
        expected = 'test_gene'
        f = FakeFeature("CDS")
        f.qualifiers['gene'] = [expected]
        f.qualifiers['protein_id'] = ['test_id']

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)

    def test_get_gene_id_no_id(self):
        "Test utils.get_gene_id() without any useable id"
        expected = 'no_tag_found'
        f = FakeFeature("CDS")

        ret = utils.get_gene_id(f)
        self.assertEqual(ret, expected)


    def test_get_aa_sequence(self):
        "Test utils.get_aa_sequence() for straightforward translation"
        expected = 'MAGIC'
        f = FakeFeature("CDS")
        f.qualifiers['translation'] = [expected]

        ret = utils.get_aa_sequence(None, f)
        self.assertEqual(expected, ret)


    def test_get_aa_sequence_stop(self):
        "Test utils.get_aa_sequence() for translation including a stop codon"
        inseq = 'MAGIC*'
        expected = 'MAGICX'
        f = FakeFeature("CDS")
        f.qualifiers['translation'] = [inseq]

        ret = utils.get_aa_sequence(None, f)
        self.assertEqual(expected, ret)


    def test_get_aa_sequence_gap(self):
        "Test utils.get_aa_sequence() for translation including a gap"
        inseq = 'MA-GIC'
        expected = 'MAGIC'
        f = FakeFeature("CDS")
        f.qualifiers['translation'] = [inseq]

        ret = utils.get_aa_sequence(None, f)
        self.assertEqual(expected, ret)


    def test_get_aa_sequence_to_stop(self):
        "Test utils.get_aa_sequence() for translation up to a stop codon"
        inseq = 'MAGIC*SEQ'
        expected = 'MAGIC'
        f = FakeFeature("CDS")
        f.qualifiers['translation'] = [inseq]

        ret = utils.get_aa_sequence(None, f, to_stop=True)
        self.assertEqual(expected, ret)
